/*
 *  qtstalker-plugin-mysql
 *
 *  Copyright (C) 2003-2007 Marco van Zwetselaar
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#include "MySQLPlugin.h"
#include "PrefDialog.h"
#include "Setting.h"
#include "Bar.h"
#include <qdir.h>
#include <qtimer.h>
#include <qdatetime.h>
#include <qsettings.h>
#include <qmessagebox.h>
#include <qstringlist.h>
#include <qinputdialog.h>
#include "configure.xpm"
#include "newchart.xpm"

/**
 * constructor
 */
MySQLPlugin::MySQLPlugin ()
{
  // plugin defaults
  pluginName = "MySQL";
  helpFile = "mysql.html";
  cancelFlag = false;

  // set base path to charts
  Config config;
  config.getData(Config::DataPath, baseChartPath);
  baseChartPath.append("/Stocks/MySQL");

  // local default
  fullReload = false;

  // load persisted database configuration
  retrieveSettings();

  // add items to the dialog
  extendGui();
}

/**
 * destructor - don't count on it to be called
 */
MySQLPlugin::~MySQLPlugin ()
{
  qtstalkerdb.close();
}

/**
 * retrieveSettings
 * 
 * Retrieve settings from Qt settings registry (~/.qt/qtstalkerrc)
 */
void MySQLPlugin::retrieveSettings() 
{
  QSettings settings;

  settings.beginGroup("/Qtstalker/MySQL plugin");
    database = settings.readEntry("/database");
    host = settings.readEntry("/host", "localhost");
    username = settings.readEntry("/username");
    password = settings.readEntry("/password");
    sqlquery = settings.readEntry("/sqlquery", 
          "SELECT day,open,high,low,close,volume FROM Quotes"
          " WHERE symbol = '$SYMBOL$' AND day > '$LASTDAY$'"
          " ORDER BY day");
  settings.endGroup();
}

/**
 * storeSettings
 * 
 * Store settings in Qt settings registry (~/.qt/qtstalkerrc)
 */
void MySQLPlugin::storeSettings()
{
  QSettings settings;
  
  settings.beginGroup("/Qtstalker/MySQL plugin");
    settings.writeEntry("/database", database);
    settings.writeEntry("/host", host);
    settings.writeEntry("/username", username);
    settings.writeEntry("/password", password);
    settings.writeEntry("/sqlquery", sqlquery);
  settings.endGroup();
}

/**
 * virtual update()
 * overrides Plugin::update()
 *
 * Called by Qtstalker in reaction to update buttonpress
 */
void MySQLPlugin::update () 
{
  qtstalkerdb.close();
  statusLog->clear();
  QTimer::singleShot(0, this, SLOT(performUpdate()));
}

/**
 * slot cancelUpdate()
 * wired from Qtstalker-provided plugin cancel button
 *
 * Called by application in reaction to 'cancel' buttonpress
 */
void MySQLPlugin::cancelUpdate ()
{
  cancelFlag = TRUE;
}


/**
 * slot performUpdate()
 *
 * Do the actual work of looping over all symbols to be updated.
 */
void MySQLPlugin::performUpdate ()
{
  if (openDatabase()) 
  {
    QString msg = fullReload ? tr("Starting full reload") : ("Starting update");
    printStatusLogMessage(msg);

    QDir dir(baseChartPath);
    if (dir.exists()) 
    {
      QStringList symbolList = dir.entryList(QDir::Files);

      if (!symbolList.isEmpty()) {

        progressBar->setTotalSteps(symbolList.count());
        progressBar->setProgress(0);

        QStringList::const_iterator iter = symbolList.begin();
        while (iter != symbolList.end() && !cancelFlag)
        {
          QString symbol = *iter++;
          updateSymbol(symbol);
          progressBar->setProgress(progressBar->progress() + 1);
        }
      }
    }
    else
    {
      QString errstr = "No symbols found to update.";
      QMessageBox::critical(0, "MySQL Plugin Error", errstr);
      QString errmsg = "MySQL Plugin Error: " + errstr;
      printStatusLogMessage(errmsg);
    }
  
    closeDatabase();
  }

  downloadComplete();

  if (!cancelFlag)
  {
    QString msg = tr("Done");
    printStatusLogMessage(msg);
  }
  cancelFlag = false;
}

/**
 * addSymbol (symbol)
 *
 * Create a new empty chart
 */
void MySQLPlugin::addSymbol (const QString& symbol)
{
  QString message;
  QString chartpath = baseChartPath;
  chartpath.append("/" + symbol);
  
  QFileInfo fi(chartpath);
  if ( !fi.exists() ) 
  {
    if ( qtstalkerdb.open(chartpath, chartIndex) )
    {
      message = "Qtstalker::MySQL::addSymbol:Could not open db: " + chartpath;
    }
    else // Dunno what exactly to do here - got it working anyway ...
    {
      QString fn = fi.fileName();
      DBIndexItem item;
      chartIndex->getIndexItem(fn, item);
      qtstalkerdb.createNewStock(); // ignore retval - seems spurious

      message = "Adding " + symbol;
      chartIndex->getIndexItem(fn, item);
      QString ncsymbol = symbol;
      item.setSymbol(ncsymbol);
      item.setTitle(ncsymbol);
      item.setQuotePlugin(pluginName);
      chartIndex->setIndexItem(fn, item);
      qtstalkerdb.close();
    }
  }
  else
  {
    message = "Skipping existing " + symbol;
  }

  printStatusLogMessage(message);
}

/**
 * updateSymbol(symbol)
 *
 * Update the quotes for given symbol.
 */
void MySQLPlugin::updateSymbol(const QString &symbol) 
{
  QString msg = "Updating " + symbol;
  printStatusLogMessage(msg);

  // open or create the new chart
  QString chartpath = baseChartPath;
  chartpath.append("/" + symbol);
  
  if (qtstalkerdb.open(chartpath, chartIndex))
  {
    QString msg = "Qtstalker::MySQL::updateSymbol:Could not open db: " + chartpath;
    printStatusLogMessage(msg);
  }
  else
  {
    QFileInfo fi(chartpath);
    QString fn = fi.fileName();

    DBIndexItem item;
    chartIndex->getIndexItem(fn, item);
  
    // determine last date present
    QDate lastdate;

    if (!fullReload) 
    {
      Bar lastbar;
      qtstalkerdb.getLastBar(lastbar);
      if (!lastbar.getEmptyFlag())
      {
  QDateTime dt;
  lastbar.getDate(dt);
  lastdate = dt.date();
      }
    }

    if (!lastdate.isValid()) 
      lastdate.setYMD(1800, 1, 1);

    // do the query
    QString sql = sqlquery;
    sql.replace("$SYMBOL$", symbol);
    sql.replace("$LASTDAY$", lastdate.toString(Qt::ISODate));

    doQuery(sql);

    qtstalkerdb.close();
  }
}

/**
 * openDatabase()
 */
bool MySQLPlugin::openDatabase()
{
  if (!(mysql_init(&mysql) 
      && mysql_real_connect(&mysql,
  host, 
  username, 
  password,
  database, 
  0     /* port */, 
  NULL  /* socket */, 
  0     /* flags */)))
  {
    QString errstr = "Could not connect to database.\n";
    errstr.append(mysql_error(&mysql));
    QMessageBox::critical(0, "Database error", errstr); 
    QString msg = "Database error: " + errstr;
    printStatusLogMessage(msg);
    return false;
  }
  
  return true;
}

/**
 * closeDatabase()
 */
void MySQLPlugin::closeDatabase()
{
  mysql_close(&mysql);
}  

/**
 * doQuery(sql,db)
 * 
 * Perform sql query and store results in Qtstalker db
 */
void MySQLPlugin::doQuery (const QString &sql)
{
  MYSQL_RES *res;
 
  if (mysql_query (&mysql, sql) == 0 
      && (res = mysql_store_result(&mysql)) != NULL) 
  {
    MYSQL_ROW row;

    bool with_oi = mysql_num_fields(res) == 7;

    while ((row = mysql_fetch_row(res)) != NULL) 
    {
      QString d = row[0];
      d = d.remove('-');
      d.append("000000");
      
      Bar bar;
      if (bar.setDate(d))
      {
        QString msg = "Bad date " + d;
        printStatusLogMessage(msg);
        continue;
      }
      
      QString open = row[1];
      QString high = row[2];
      QString low = row[3];
      QString close = row[4];
      QString volume = row[5];
      
      bar.setOpen(open.toDouble());
      bar.setHigh(high.toDouble());
      bar.setLow(low.toDouble());
      bar.setClose(close.toDouble());
      bar.setVolume(volume.toDouble());

      if (with_oi) {
        QString oi = row[6];
        bar.setOI(oi.toInt());
      }

      qtstalkerdb.setBar(bar);

      emit signalWakeup();
    }
   
    mysql_free_result(res);
  }
  else 
  {
    QString errstr = "Database query failed.\n";
    errstr.append(mysql_error(&mysql));
    QMessageBox::critical(0, "Database Query problem", errstr);
    QString errmsg = "Database Query problem: " + errstr;
    printStatusLogMessage(errmsg);
  }
}

/**
 * extendGui ()
 *
 * Adds plugin specific buttons and toggles to the update dialog
 */
void MySQLPlugin::extendGui() 
{
  setCaption(tr("MySQL Quote Update"));

  // connect existing cancel button to our cancelUpdate slot
  QString cancelBtnName = "cancelDownload";
  QObject::connect(toolbar->getButton(cancelBtnName), SIGNAL(clicked()), this, SLOT(cancelUpdate()));
  toolbar->setButtonStatus(cancelBtnName, FALSE);

  // add a button to add symbols
  QString name = "addSymbolsButton";
  QString hint = tr("Add symbols to import");
  toolbar->addButton(name, newchart_icon, hint);
  QObject::connect(toolbar->getButton(name), SIGNAL(clicked()), this, SLOT(addSymbolsDialog()));

  // add a button to configure the MySQL plugin
  name = "editSettingsButton";
  hint = tr("Configure plugin settings");
  toolbar->addButton(name, configure_icon, hint);
  QObject::connect(toolbar->getButton(name), SIGNAL(clicked()), this, SLOT(editSettingsDialog()));

  // add a checkbox to toggle full reload
  QCheckBox *check = new QCheckBox(tr("Full reload"), baseWidget);
  check->setChecked(fullReload);
  grid->addWidget(check, 0, 0);
  QObject::connect(check, SIGNAL(toggled(bool)), this, SLOT(fullReloadToggle(bool)));
} 

/**
 * slot fullReloadToggle (bool)
 * triggered by the checkbox
 */
void MySQLPlugin::fullReloadToggle (bool b)
{
  fullReload = b;
}

/**
 * slot addSymbolsDialog ()
 * triggered by button press
 */
void MySQLPlugin::addSymbolsDialog ()
{
  bool okPressed = false;
  QString newSymbols = QInputDialog::getText(
      tr("Add symbols"), tr("Enter symbol names separated by spaces"),
      QLineEdit::Normal, QString::null, &okPressed, baseWidget);

  if (okPressed)
  {
    QStringList symbolList = QStringList::split(' ', newSymbols);
    if (!symbolList.isEmpty()) 
    {
      QString chartpath = baseChartPath;
      QDir dir(chartpath);
      if (dir.exists() || dir.mkdir(chartpath))
      {
        statusLog->clear();
        QString msg = tr("Adding symbols ...");
        printStatusLogMessage(msg);

        progressBar->setTotalSteps(symbolList.count());
        progressBar->setProgress(0);

        QStringList::const_iterator iter = symbolList.begin();
        while (iter != symbolList.end())
        {
          QString symbol = *iter++;
          addSymbol (symbol);
          progressBar->setProgress(progressBar->progress() + 1);
        }

        msg = tr("Done");
        printStatusLogMessage(msg);
      }
      else
      {
        QString errstr = "Qtstalker::MySQL::addSymbols:unable to create directory: ";
        errstr.append(chartpath);
        QMessageBox::critical(0, "MySQL Plugin Error", errstr);
        QString errmsg = "MySQL Plugin Error: " + errstr;
        printStatusLogMessage(errmsg);
      }
    }
  }
}

/**
 * slot editSettingsDialog ()
 * triggered by button press
 */
void MySQLPlugin::editSettingsDialog ()
{
  PrefDialog *dialog = new PrefDialog(baseWidget);
  dialog->setCaption(tr("MySQL Plugin Configuration"));
  QString s = tr("Database Settings");
  dialog->createPage (s);
  dialog->setHelpFile(helpFile);

  QString s2;
  s2 = tr("Database");
  dialog->addTextItem(s2, s, database);
  s2 = tr("Host");
  dialog->addTextItem(s2, s, host);
  s2 = tr("Username");
  dialog->addTextItem(s2, s, username);
  s2 = tr("Password");
  dialog->addTextItem(s2, s, password);
  s2 = tr("SQL Query");
  dialog->addTextItem(s2, s, sqlquery);
  
  int rc = dialog->exec();
  
  if (rc == QDialog::Accepted)
  {
    s = tr("Database");
    dialog->getText(s, database);
    s = tr("Host");
    dialog->getText(s, host);
    s = tr("Username");
    dialog->getText(s, username);
    s = tr("Password");
    dialog->getText(s, password);
    s = tr("SQL Query");
    dialog->getText(s, sqlquery);

    storeSettings();
  }
  
  delete dialog;
}

/**
 * create()
 *
 * Return new instance of this plugin.
 */
QuotePlugin * createQuotePlugin ()
{
  MySQLPlugin *o = new MySQLPlugin;
  return ((QuotePlugin *) o);
}


